SystemOrganization addCategory: #'Goya-VM'!
SystemOrganization addCategory: #'Goya-Browser'!

!Interpreter methodsFor: '*Goya-VM' stamp: 'kwl 11/6/2005 10:52'!
commonSend
	"Send a message, starting lookup with the receiver's class."
	"Assume: messageSelector and argumentCount have been set, and that 
	the receiver and arguments have been pushed onto the stack,"
	"Note: This method is inlined into the interpreter dispatch loop."
	self sharedCodeNamed: 'commonSend' inCase: 131.
	"kwl 11/2005 - instrumenting - internalFindNewMethod not used here."
	self findNewMethodInClass: lkupClass.
	self internalExecuteNewMethod.
	self fetchNextBytecode! !

!Interpreter methodsFor: '*Goya-VM' stamp: 'kwl 11/19/2005 15:59'!
findNewMethodInClass: class 
	"Find the compiled method to be run when the current 
	messageSelector is sent to the given class, setting the values 
	of 'newMethod' and 'primitiveIndex'."
	| ok callerIndex callerAssoc calleeIndex calleeAssoc temp |
	self inline: false.
	ok _ self lookupInMethodCacheSel: messageSelector class: class.
	ok
		ifFalse: ["entry was not found in the cache; look it up the hard way "
			self lookupMethodInClass: class.
			lkupClass _ class.
			self addNewMethodToCache].
	"kwl 11/2005 - instrumenting"
	((callerIndex _ self literalCountOf: method) > 0
			and: [(calleeIndex _ self literalCountOf: newMethod) > 0])
		ifTrue: [((self isIntegerObject: (callerAssoc _ self literal: callerIndex - 1 ofMethod: method))
					or: [self isIntegerObject: (calleeAssoc _ self literal: calleeIndex - 1 ofMethod: newMethod)])
				ifTrue: [^ nil].
			((callerIndex _ (self baseHeader: callerAssoc)
								>> 12 bitAnd: 31) >= 27
					and: [(calleeIndex _ (self baseHeader: calleeAssoc)
									>> 12 bitAnd: 31) >= 27])
				ifTrue: ["update"
					temp _ (self fetchPointer: ValueIndex + 1 ofObject: callerAssoc)
								bitOr: (1 << (32 - calleeIndex)).
					self
						storePointer: ValueIndex + 1 ofObject: callerAssoc withValue: temp.
					temp _ (self fetchPointer: ValueIndex + 2 ofObject: calleeAssoc)
								bitOr: (1 << (32 - callerIndex)).
					self
						storePointer: ValueIndex + 2 ofObject: calleeAssoc withValue: temp
				]
			]! !

Browser subclass: #GoyaBrowser
	instanceVariableNames: 'classesSet ccIndex colorMask'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Goya-Browser'!

!GoyaBrowser commentStamp: '<historical>' prior: 0!
I show only methods (and their respective classes (and their respective system categories)) which where executed in the running image, filtered by the color or the caller and/or callee.

See the class side of InstrumentingAssociation for instrumenting CompiledMethods to be colored by the VM.!

!GoyaBrowser methodsFor: 'initialize-release' stamp: 'kwl 11/20/2005 02:13'!
systemOrganizer: aSystemOrganizer
	"Initialize the receiver as a perspective on the system organizer, 
	aSystemOrganizer. Typically there is only one--the system variable 
	SystemOrganization."

	super systemOrganizer: (SystemOrganizer defaultList: Array new).
	self pvtRefreshLists! !

!GoyaBrowser methodsFor: 'initialize-release' stamp: 'kwl 11/20/2005 02:32'!
initialize
	"Initialize the state of the receiver."

	super initialize.
	classesSet _ IdentitySet new.
	ccIndex _ 30.
	colorMask _ 1! !

!GoyaBrowser methodsFor: 'system category functions' stamp: 'kwl 11/20/2005 02:14'!
updateSystemCategories
	"The class categories were changed."

	self pvtRefreshLists.
	super updateSystemCategories! !

!GoyaBrowser class methodsFor: 'instance creation' stamp: 'kwl 11/19/2005 21:54'!
prototypicalToolWindow
	"Answer an example of myself seen in a tool window, for the benefit of parts-launching tools"

	| aWindow |
	aWindow _ self new openAsMorphGoyaEditing: nil.
	aWindow setLabel: 'Goya Browser'; applyModelExtent.
	^ aWindow! !

!GoyaBrowser methodsFor: 'initialize-release' stamp: 'kwl 11/19/2005 21:57'!
defaultBrowserTitle
	^ 'Goya Browser'! !

!GoyaBrowser class methodsFor: 'class initialization' stamp: 'kwl 11/19/2005 21:15'!
initialize
	"GoyaBrowser initialize"

	self registerInFlapsRegistry.	! !

!GoyaBrowser class methodsFor: 'class initialization' stamp: 'kwl 11/19/2005 21:14'!
registerInFlapsRegistry
	"Register the receiver in the system's flaps registry"
	self environment
		at: #Flaps
		ifPresent: [:cl | cl registerQuad: #(#GoyaBrowser #prototypicalToolWindow 'GoyaBrowser' 'A GoyaBrowser is a tool that allows you to view the active code of the active classes in the system' ) 
						forFlapNamed: 'Tools']! !

!GoyaBrowser methodsFor: 'message list' stamp: 'kwl 11/20/2005 02:30'!
messageList
	"Answer an Array of the message selectors of the currently selected  
	class."
	| list |
	(list _ super messageList) isEmpty
		ifTrue: [^ list].
	^ Array
		streamContents: [:stream | self selectedClassOrMetaClass
				selectorsAndMethodsDo: [:selector :method | | assoc callerColor | ((method isQuick not
								and: [method literals size > 0])
							and: [(assoc _ method literals last) class inheritsFrom: InstrumentingAssociation])
						ifTrue: [callerColor _ assoc callerBits.
							(assoc class indexIfCompact = ccIndex
									and: [(callerColor bitAnd: colorMask)
											= colorMask])
								ifTrue: [stream nextPut: selector]]]]! !

!GoyaBrowser methodsFor: 'initialize-release' stamp: 'kwl 11/19/2005 21:54'!
openAsMorphGoyaEditing: editString
	"Create a pluggable version of all the morphs for a GoyaBrowser in Morphic"
	| window hSepFrac |

	hSepFrac _ 0.4.
	window _ (SystemWindow labelled: 'later') model: self.

"The method SystemWindow>>addMorph:fullFrame: checks scrollBarsOnRight, then adds the morph at the back if true, otherwise it is added in front. But flopout hScrollbars need the lowerpanes to be behind the upper ones in the draw order. Hence the value of scrollBarsOnRight affects the order in which the lowerpanes are added. "
	Preferences scrollBarsOnRight ifFalse:
		[self 
			addLowerPanesTo: window 
			at: (0@hSepFrac corner: 1@1) 
			with: editString].
		
	window 
		addMorph: self buildMorphicSystemCatList
		frame: (0@0 corner: 1.0/3@hSepFrac).
	self 
		addClassAndSwitchesTo: window 
		at: (1.0/3@0 corner: 2.0/3@hSepFrac)
		plus: 0.
	window addMorph: self buildMorphicMessageList
		frame: (2.0/3@0 extent: 1.0/3@hSepFrac).

	Preferences scrollBarsOnRight ifTrue:
		[self 
			addLowerPanesTo: window 
			at: (0@hSepFrac corner: 1@1) 
			with: editString].

	window setUpdatablePanesFrom: #(systemCategoryList classList messageList).
	^ window
! !

!GoyaBrowser methodsFor: 'class list' stamp: 'kwl 11/20/2005 01:59'!
classList
	"Answer an array of the class names of the selected category. Answer an 
	empty array if no selection exists."

	systemCategoryListIndex = 0
		ifTrue: [^Array new]
		ifFalse: [^(systemOrganizer listAtCategoryNumber: systemCategoryListIndex) select: [:each | classesSet includes: each]]! !

!GoyaBrowser class methodsFor: 'window color' stamp: 'kwl 11/19/2005 21:32'!
windowColorSpecification
	"Answer a WindowColorSpec object that declares my preference"

	^ WindowColorSpec classSymbol: self name  wording: 'GoyaBrowser' brightColor: #lightYellow pastelColor: #paleYellow helpMessage: 'The Goya "browser" tool that allows you to browse through the active code in the system'! !

!GoyaBrowser methodsFor: 'private' stamp: 'kwl 11/20/2005 03:45'!
pvtRefreshLists
	classesSet _ IdentitySet new.
	systemOrganizer _ SystemOrganizer defaultList: Array new.
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | assoc callerColor calleeColor className | ((method isQuick not
								and: [method literals size > 0])
							and: [(assoc := method literals last) class inheritsFrom: InstrumentingAssociation])
						ifTrue: [callerColor := assoc callerBits.
							calleeColor := assoc calleeBits.
							(assoc class indexIfCompact = ccIndex
									and: [(callerColor bitAnd: colorMask)
											= colorMask])
								ifTrue: [classesSet add: (className := class theNonMetaClass name).
									systemOrganizer
										classify: className
										under: (SystemOrganization categoryOfElement: className)]]]].
		systemOrganizer removeEmptyCategories! !

Association subclass: #InstrumentingAssociation
	instanceVariableNames: 'callerBits calleeBits'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Goya-VM'!

!InstrumentingAssociation commentStamp: '<historical>' prior: 0!
Subinstances of InstrumentingAssociation (which are compact classes), are objects which can be detected by the VM in O(1) time and make their living in the last literal of CompiledMethods. They have the same structure as the association used by super sends (i.e. the value field stored the CompiledMethod's defining class) but they also store the CompiledMethod's defining selector (in the key field).

Because other classes can be compact too, we require that InstrumentingAssociation subclasses have the largest index in the #compactClassesArray.

Possible useful expressions for doIt or printIt: see InstrumentingAssociationTest.

Structure:
 callerBits	SmallInteger -- updated by the VM, sets the (32-index)'th bit of the callee's methodClass association with compact class index
 calleeBits	SmallInteger -- updated by the VM, sets the (32-index)'th bit of the caller's methodClass association with compact class index

Further useful comments about the general approach of this implementation: see #compiledMethod:selector:class: for how to implant a subinstance of InstrumentingAssociation in a CompiledMethod.!

InstrumentingAssociation subclass: #Instrumenting27Association
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Goya-VM'!

!Instrumenting27Association commentStamp: 'kwl 11/19/2005 11:05' prior: 0!
See InstrumentingAssociation class comment.!

!Instrumenting27Association class methodsFor: 'class initialization' stamp: 'kwl 11/18/2005 13:43'!
initialize
	"Instrumenting27Association initialize"
	"This cannot be done more that once!!"
	Instrumenting27Association becomeCompactSimplyAt: Object environment compactClassesArray size - 4! !

InstrumentingAssociation subclass: #Instrumenting31Association
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Goya-VM'!

!Instrumenting31Association commentStamp: '<historical>' prior: 0!
See InstrumentingAssociation class comment.!

!Instrumenting31Association class methodsFor: 'class initialization' stamp: 'kwl 11/19/2005 16:00'!
initialize
	"Instrumenting31Association initialize"
	"This cannot be done more that once!!"
	Instrumenting31Association becomeCompactSimplyAt: Object environment compactClassesArray size - 0! !

InstrumentingAssociation subclass: #Instrumenting30Association
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Goya-VM'!

!Instrumenting30Association commentStamp: '<historical>' prior: 0!
See InstrumentingAssociation class comment.!

!Instrumenting30Association class methodsFor: 'class initialization' stamp: 'kwl 11/18/2005 13:36'!
initialize
	"Instrumenting30Association initialize"
	"This cannot be done more that once!!"
	Instrumenting30Association becomeCompactSimplyAt: Object environment compactClassesArray size - 1! !

!InstrumentingAssociation methodsFor: 'initialize-release' stamp: 'kwl 11/19/2005 17:29'!
compiledMethod: aMethod selector: aSelector class: aClass 
	"Answer a CompiledMethod with the receiver implanted as the last literal."
	| numLits literals newHeader newMethod j |
	self key: aSelector value: aClass.
	numLits _ (literals _ aMethod literals) size.
	(numLits > 1
			and: [self = (literals at: numLits)])
		ifTrue: ["just replace with the receiver"
			aMethod literalAt: numLits put: self.
			^ aMethod].
	"implant the receiver in this method"
	(aMethod sendsToSuper
			or: [numLits > 1 and: [(literals at: numLits) class inheritsFrom: InstrumentingAssociation]])
		ifTrue: [literals _ literals copyFrom: 1 to: literals size - 1].
	literals _ literals copyWith: self.
	literals size > 255
		ifTrue: [^ self error: 'Too many literals referenced'].
	newHeader _ (aMethod header bitClear: (255 bitShift: 9)) bitOr: (literals size bitShift: 9).
	newMethod _ aMethod class newMethod: aMethod size - aMethod initialPC + 1 header: newHeader.
	1 to: literals size
		do: [:i | newMethod literalAt: i put: (literals at: i)].
	j _ newMethod initialPC.
	aMethod initialPC to: aMethod size
		do: [:i | newMethod at: j put: (aMethod at: i). j _ j + 1].
	^ newMethod! !

!InstrumentingAssociation methodsFor: 'accessing' stamp: 'kwl 11/19/2005 11:48'!
calleeBits
	^ calleeBits! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/19/2005 21:03'!
preInstrumentAllPrimitives
	"InstrumentingAssociation preInstrumentAllPrimitives"
	"Color primitive methods 30 and others methods 31."
	"Check with 'InstrumentingAssociation reportIndex: 30 caller: 1' for primitives which actually fall back to their method's bytecode"
	"Check with 'InstrumentingAssociation reportIndex: 31 callee: 2' for what these fallen entities actually send"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | newMethod | (newMethod := (method primitive = 0
									ifTrue: [Instrumenting31Association]
									ifFalse: [Instrumenting30Association]) new
									compiledMethod: method
									selector: selector
									class: class) == method
						ifFalse: [class addSelectorSilently: selector withMethod: newMethod]]]! !

!InstrumentingAssociation methodsFor: 'accessing' stamp: 'kwl 11/19/2005 11:48'!
callerBits
	^ callerBits! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/19/2005 20:32'!
preInstrumentAllConstructors
	"InstrumentingAssociation preInstrumentAllConstructors"
	"Color creation of new instances black and others methods white."
	"Check with 'InstrumentingAssociation reportIndex: 30 caller: 1' for actual senders (many)"
	"Check with 'InstrumentingAssociation reportIndex: 31 caller: 2' for used implementors (few)"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | newMethod | (newMethod := (((method primitive between: 70 and: 71)
										or: [#(#basicNew #basicNew: #new #new: ) includes: selector])
									ifTrue: [Instrumenting31Association]
									ifFalse: [Instrumenting30Association]) new
									compiledMethod: method
									selector: selector
									class: class) == method
						ifFalse: [class addSelectorSilently: selector withMethod: newMethod]]]! !

!InstrumentingAssociation methodsFor: 'initialize-release' stamp: 'kwl 11/12/2005 16:47'!
initialize
	"Answer an initialized instance of the receiver"
	callerBits _ calleeBits _ 0! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/20/2005 03:47'!
preInstrumentFileStreams
	"InstrumentingAssociation preInstrumentFileStreams"
	"Color FileStream 31 and others methods 30."
	"Check with 'InstrumentingAssociation reportIndex: 30 caller: 1' for users of the FileStream behavior"
	"Check with 'InstrumentingAssociation reportIndex: 31 caller: 2' for used implementors"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | newMethod | (newMethod := ((class theNonMetaClass includesBehavior: FileStream)
									ifTrue: [Instrumenting31Association]
									ifFalse: [Instrumenting30Association]) new
									compiledMethod: method
									selector: selector
									class: class) == method
						ifFalse: [class addSelectorSilently: selector withMethod: newMethod]]]! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/20/2005 03:40'!
preInstrumentMorphs
	"InstrumentingAssociation preInstrumentMorphs"
	"Color Morphs 31 and others methods 30."
	"Check with 'InstrumentingAssociation reportIndex: 30 caller: 1' for users of the Morph behavior (few)"
	"Check with 'InstrumentingAssociation reportIndex: 31 caller: 2' for used implementors (many)"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | newMethod | (newMethod := ((class theNonMetaClass includesBehavior: Morph)
									ifTrue: [Instrumenting31Association]
									ifFalse: [Instrumenting30Association]) new
									compiledMethod: method
									selector: selector
									class: class) == method
						ifFalse: [class addSelectorSilently: selector withMethod: newMethod]]]! !

InstrumentingAssociation subclass: #Instrumenting29Association
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Goya-VM'!

!Instrumenting29Association commentStamp: '<historical>' prior: 0!
See InstrumentingAssociation class comment.!

!Instrumenting29Association class methodsFor: 'class initialization' stamp: 'kwl 11/18/2005 13:36'!
initialize
	"Instrumenting29Association initialize"
	"This cannot be done more that once!!"
	Instrumenting29Association becomeCompactSimplyAt: Object environment compactClassesArray size - 2! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/19/2005 20:56'!
reportIndex: ccIndex callee: mask 
	"InstrumentingAssociation reportIndex: 31 callee: 2"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | assoc callerColor calleeColor | ((method isQuick not
								and: [method literals size > 0])
							and: [(assoc _ method literals last) class inheritsFrom: InstrumentingAssociation])
						ifTrue: [callerColor _ assoc callerBits.
							calleeColor _ assoc calleeBits.
							(assoc class indexIfCompact = ccIndex
									and: [(calleeColor bitAnd: mask)
											= mask])
								ifTrue: [Transcript cr; show: class name; space; print: assoc class indexIfCompact; space; show: selector; space; print: callerColor; space; print: calleeColor; endEntry]]]]! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/19/2005 20:56'!
reportIndex: ccIndex caller: mask
	"InstrumentingAssociation reportIndex: 30 caller: 1"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | assoc callerColor calleeColor | ((method isQuick not
								and: [method literals size > 0])
							and: [(assoc _ method literals last) class inheritsFrom: InstrumentingAssociation])
						ifTrue: [callerColor _ assoc callerBits.
							calleeColor _ assoc calleeBits.
							(assoc class indexIfCompact = ccIndex
									and: [(callerColor bitAnd: mask)
											= mask])
								ifTrue: [Transcript cr; show: class name; space; print: assoc class indexIfCompact; space; show: selector; space; print: callerColor; space; print: calleeColor; endEntry]]]]! !

InstrumentingAssociation subclass: #Instrumenting28Association
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Goya-VM'!

!Instrumenting28Association commentStamp: 'kwl 11/19/2005 11:05' prior: 0!
See InstrumentingAssociation class comment.!

!Instrumenting28Association class methodsFor: 'class initialization' stamp: 'kwl 11/18/2005 13:42'!
initialize
	"Instrumenting28Association initialize"
	"This cannot be done more that once!!"
	Instrumenting28Association becomeCompactSimplyAt: Object environment compactClassesArray size - 3! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/20/2005 03:47'!
preInstrumentBrowser
	"InstrumentingAssociation preInstrumentBrowser"
	"Color Morphs 31 and others methods 30."
	"Check with 'InstrumentingAssociation reportIndex: 30 caller: 1' for users of the Browser behavior (few)"
	"Check with 'InstrumentingAssociation reportIndex: 31 caller: 2' for used implementors (many)"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | newMethod | (newMethod := ((class theNonMetaClass includesBehavior: Browser)
									ifTrue: [Instrumenting31Association]
									ifFalse: [Instrumenting30Association]) new
									compiledMethod: method
									selector: selector
									class: class) == method
						ifFalse: [class addSelectorSilently: selector withMethod: newMethod]]]! !

!InstrumentingAssociation class methodsFor: 'pre instrumenting' stamp: 'kwl 11/20/2005 03:17'!
preInstrumentSystemCompiler
	"InstrumentingAssociation preInstrumentSystemCompiler"
	"Color the System-Compiler category with 30 and others methods 31."
	"Check with 'InstrumentingAssociation reportIndex: 30 caller: 1' for active methods in the System-Compiler category"
	self systemNavigation
		allBehaviorsDo: [:class | class
				selectorsAndMethodsDo: [:selector :method | | newMethod | (newMethod := ((SystemOrganization categoryOfElement: class theNonMetaClass name) = 'System-Compiler'
									ifFalse: [Instrumenting31Association]
									ifTrue: [Instrumenting30Association]) new
									compiledMethod: method
									selector: selector
									class: class) == method
						ifFalse: [class addSelectorSilently: selector withMethod: newMethod]]]! !
GoyaBrowser initialize!
Instrumenting27Association initialize!
Instrumenting31Association initialize!
Instrumenting30Association initialize!
Instrumenting29Association initialize!
Instrumenting28Association initialize!
